/* 
 * $TSUKUBA_Release: Omni OpenMP/XcalableMP Compiler Software Version 0.6.0 (alpha) $
 * $TSUKUBA_Copyright:
 *  Copyright (C) 2010-2011 University of Tsukuba, 
 *  	      2012  University of Tsukuba and Riken AICS
 *  
 *  This software is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License version
 *  2.1 published by the Free Software Foundation.
 *  
 *  Please check the Copyright and License information in the files named
 *  COPYRIGHT and LICENSE under the top  directory of the Omni Compiler
 *  Software release kit.
 *  
 *  * The specification of XcalableMP has been designed by the XcalableMP
 *    Specification Working Group (http://www.xcalablemp.org/).
 *  
 *  * The development of this software was partially supported by "Seamless and
 *    Highly-productive Parallel Programming Environment for
 *    High-performance computing" project funded by Ministry of Education,
 *    Culture, Sports, Science and Technology, Japan.
 *  $
 */
#include <sys/param.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>

#include "c-expr.h"
#include "c-pragma.h"
#include "c-parser.h"
#include "c-const.h"
#include "c-option.h"
#include "c-acc.h"

/*
 * <ACCPragma> <string> directive_name </string> 
 *             [clauses] [body] </OMPPragma>
 * [clauses] = <list> [clause] </list>
 *   C_Front: (direcive clause1 clause2 ... )
 *
 * [data_clause] = 
 *     <list> <string> [data_clause_name] </string> [name_list] </list>
 *     [data_clause_name] = ACC_DATA_PRIVATE|
 *              ACC_DATA_FIRSTPRIVATE|ACC_DATA_LASTPRIVATE|
 *		ACC_DATA_COPYIN|
 *               ACC_DATA_REDUCTION_***
 *     [name_list] = <list> variable ... </list>
 *  C_Front: (data_clause_name (LIST ident ... ))
 * 
 * [default_clause] = 
 *      <list> <string> ACC_DATA_DEFAULT </string> 
 *           <string> ACC_DEFAULT_*** </string> </list>
 *  C_Front: (ACC_DATA_DEFAULT (ACC_DEFAULT_*** null))
 * 
 * [if_clause] = <list> <string> ACC_DIR_IF </string> cond_expr </list>
 *  C_Front: (ACC_DIR_IF cond_expr)
 *
 * [schedule_clause] = 
 *       <list> <string> ACC_DIR_SCHEDULE </string>
 *           <list> <string> ACC_SCHED_*** </string> expr </list> </list>
 *  C_Front: (ACC_DIR_SCHEDULE (ACC_SCHED_*** expr))
 *
 * [ordered_clause] = <list> <string> ACC_DIR_ORDERED </strign> null </list>
 *  C_Front: (ACC_DIR_ORDERED null) 
 *
 * [nowait_clause] = <list> <string> ACC_DIR_NOWAIT </strign> null </list>
 *  C_Front: (ACC_DIR_NOWAIT null) 
 *
 * [num_threads_clause] = 
 *    <list> <string> ACC_DIR_NUM_THREADS </strign> expr </list>
 *  C_Front: (ACC_DIR_NUM_THREADS expr) 
 *
 */

static int parse_ACC_pragma(void);
static CExpr* parse_ACC_clauses(void);
static CExpr* parse_ACC_namelist(void);
static CExpr* parse_ACC_reduction_namelist(int *r);
static CExpr* parse_ACC_clause_arg(void);

#define ACC_PG_LIST(pg,args) _omp_pg_list(pg,args)

static CExpr* _omp_pg_list(int omp_code,CExpr* args)
{
  CExprOfList *lp;
  lp = allocExprOfList1(EC_UNDEF,args);
  lp->e_aux = omp_code;
  return (CExpr *)lp;
}

#define EMPTY_LIST (CExpr *)allocExprOfList(EC_UNDEF)

#ifdef not
static expv compile_ACC_SECTIONS_statement(expr x);
static void compile_ACC_pragma_clause(expr x, int pragma, int is_parallel, expv *pc, expv *dc);
static void compile_ACC_name_list _ANSI_ARGS_((expr x));
#endif

static int pg_ACC_pragma;
CExpr* pg_ACC_list;

/*
 * for OpenACC directives
 */
CExpr*
lexParsePragmaACC(char *p, int *token) // p is buffer
{
  //skip pragma[space]omp[space]*
  p = lexSkipSpace(lexSkipWordP(lexSkipSpace(lexSkipWord(lexSkipSpace(p)))));

  pg_cp = p; // set the pointer

  *token = parse_ACC_pragma();

  if(pg_ACC_list == NULL) pg_ACC_list = EMPTY_LIST;
  ((CExprOfList *)pg_ACC_list)->e_aux = pg_ACC_pragma;
  
  return pg_ACC_list;
}

int parse_ACC_pragma()
{
    int ret = PRAGMA_PREFIX; /* default */
    pg_ACC_list = NULL;

    pg_get_token();
    if(pg_tok != PG_IDENT) goto syntax_err;

    if(PG_IS_IDENT("parallel")){
	pg_get_token();
	if(pg_tok == PG_IDENT){
	    if(PG_IS_IDENT("loop")){	/* parallel for */
		pg_ACC_pragma = ACC_PARALLEL_LOOP;
		pg_get_token();
		if((pg_ACC_list = parse_ACC_clauses()) == NULL) 
		    goto syntax_err;
		goto chk_end;
	    }
	}
	pg_ACC_pragma = ACC_PARALLEL;
	if((pg_ACC_list = parse_ACC_clauses()) == NULL) goto syntax_err;
	goto chk_end;
    }
  
    if(PG_IS_IDENT("kernels")){
	pg_get_token();
	if(pg_tok == PG_IDENT){
	    if(PG_IS_IDENT("loop")){	/* parallel for */
		pg_ACC_pragma = ACC_KERNELS_LOOP;
		pg_get_token();
		if((pg_ACC_list = parse_ACC_clauses()) == NULL) 
		    goto syntax_err;
		goto chk_end;
	    }
	}
	pg_ACC_pragma = ACC_KERNELS;
	if((pg_ACC_list = parse_ACC_clauses()) == NULL) goto syntax_err;
	goto chk_end;
    }
  
    if(PG_IS_IDENT("loop")){
	pg_ACC_pragma = ACC_LOOP;
	pg_get_token();
	if((pg_ACC_list = parse_ACC_clauses()) == NULL) goto syntax_err;
	goto chk_end;
    }

    if(PG_IS_IDENT("data")){
	pg_ACC_pragma = ACC_DATA;
	pg_get_token();
	if((pg_ACC_list = parse_ACC_clauses()) == NULL) goto syntax_err;
	goto chk_end;
    }

    if(PG_IS_IDENT("host_data")){
	pg_ACC_pragma = ACC_HOST_DATA;
	pg_get_token();
	if((pg_ACC_list = parse_ACC_clauses()) == NULL)  goto syntax_err;
	goto chk_end;
    }

    if(PG_IS_IDENT("declare")){
	pg_ACC_pragma = ACC_DECLARE;
	pg_get_token();
	if((pg_ACC_list = parse_ACC_clauses()) == NULL)  goto syntax_err;
	goto chk_end;
    }

    if(PG_IS_IDENT("update")){
	pg_ACC_pragma = ACC_UPDATE;
	pg_get_token();
	if((pg_ACC_list = parse_ACC_clauses()) == NULL)  goto syntax_err;
	ret= PRAGMA_EXEC;
	goto chk_end;
    }

    if(PG_IS_IDENT("cache")){
	pg_ACC_pragma = ACC_CACHE;
	pg_get_token();
	if(pg_tok == '('){
	    if((pg_ACC_list = parse_ACC_namelist()) == NULL) goto syntax_err;
	} else pg_ACC_list = NULL;
	ret= PRAGMA_EXEC;
	goto chk_end;
    }

    if(PG_IS_IDENT("wait")){
	pg_ACC_pragma = ACC_WAIT;
	pg_get_token();
	if(pg_tok == '('){
	    if((pg_ACC_list = parse_ACC_namelist()) == NULL) goto syntax_err;
	} else pg_ACC_list = NULL;
	ret= PRAGMA_EXEC;
	goto chk_end;
    }

    addError(NULL,"ACC: unknown ACC directive, '%s'",pg_tok_buf);
  syntax_err:
    return 0;

  chk_end:
    if(pg_tok != 0) addError(NULL,"ACC: extra arguments for ACC directive");
    return ret;
}

static CExpr* parse_ACC_clauses()
{
  CExpr *args,*v,*c;
  int r = 0;

  args = EMPTY_LIST;

  while(pg_tok == PG_IDENT){
      if(PG_IS_IDENT("private")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_PRIVATE,v);
      } else if(PG_IS_IDENT("firstprivate")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_FIRSTPRIVATE,v);
      } else if(PG_IS_IDENT("lastprivate")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_LASTPRIVATE,v);
      } else if(PG_IS_IDENT("copyin")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_COPYIN,v);
      } else if(PG_IS_IDENT("copyout")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_COPYOUT,v);
      } else if(PG_IS_IDENT("copy")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_COPY,v);
      } else if(PG_IS_IDENT("create")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_CREATE,v);
      } else if(PG_IS_IDENT("present")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_PRESENT,v);
      } else if(PG_IS_IDENT("present_or_copy")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_PRESENT_OR_COPY,v);
      } else if(PG_IS_IDENT("present_or_copyin")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_PRESENT_OR_COPYIN,v);
      } else if(PG_IS_IDENT("present_or_copyout")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_PRESENT_OR_COPYOUT,v);
      } else if(PG_IS_IDENT("present_or_create")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_PRESENT_OR_CREATE,v);
      } else if(PG_IS_IDENT("deviceptr")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_DEVICEPTR,v);
      } else if(PG_IS_IDENT("host")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_HOST,v);
      } else if(PG_IS_IDENT("device")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_DEVICE,v);
      } else if(PG_IS_IDENT("use_device")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_USE_DEVICE,v);
      } else if(PG_IS_IDENT("device_resident")){
	  pg_get_token();
	  if((v = parse_ACC_namelist()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_DEV_RESIDENT,v);
      } else if(PG_IS_IDENT("reduction")){
	  pg_get_token();
	  if((v = parse_ACC_reduction_namelist(&r)) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(r,v);
      } else if(PG_IS_IDENT("if")){  // arg
	  pg_get_token();
	  if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_IF,v);
      } else if(PG_IS_IDENT("async")){  // arg
	  pg_get_token();
	  if(pg_tok != '(') v = NULL;
	    else if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	    c = ACC_PG_LIST(ACC_ASYNC,v);
      } else if(PG_IS_IDENT("gang")){
	  pg_get_token();
	  if(pg_tok != '(') v = NULL;
	  else if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	  c = ACC_PG_LIST(ACC_GANG,v);
      } else if(PG_IS_IDENT("num_gangs")){
	    pg_get_token();
	    if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	    c = ACC_PG_LIST(ACC_NUM_GANGS,v);
	} else if(PG_IS_IDENT("worker")){
	    pg_get_token();
	    if(pg_tok != '(') v = NULL;
	    else if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	    c = ACC_PG_LIST(ACC_WORKER,v);
	} else if(PG_IS_IDENT("num_workers")){
	    pg_get_token();
	    if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	    pg_get_token();
	    c = ACC_PG_LIST(ACC_NUM_WORKERS,v);
	} else if(PG_IS_IDENT("vector")){
	    pg_get_token();
	    if(pg_tok != '(') v = NULL;
	    else if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	    c = ACC_PG_LIST(ACC_VECTOR,v);
	} else if(PG_IS_IDENT("vector_length")){
	    pg_get_token();
	    if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	    c = ACC_PG_LIST(ACC_VECT_LEN,v);
	} else if(PG_IS_IDENT("collapse")){
	    pg_get_token();
	    if((v = parse_ACC_clause_arg()) == NULL) goto syntax_err;
	    c = ACC_PG_LIST(ACC_COLLAPSE,v);
	} else if(PG_IS_IDENT("seq")){
	    pg_get_token();
	    c = ACC_PG_LIST(ACC_SEQ,NULL);
	} else if(PG_IS_IDENT("independent")){
	    pg_get_token();
	    c = ACC_PG_LIST(ACC_INDEPENDENT,NULL);
	} else {
	  addError(NULL,"unknown ACC directive clause '%s'",pg_tok_buf);
	    goto syntax_err;
	}
	args = exprListAdd(args,c);
    }
    return args;
 syntax_err:
    addError(NULL,"ACC: syntax error in ACC pragma clause");
    return NULL;
}

static CExpr* parse_ACC_namelist()
{
    CExpr* args;

    args = EMPTY_LIST;
    if(pg_tok != '(') {
      addError(NULL,"ACC: ACC directive clause requires name list");
	return NULL;
    }
    pg_get_token();
 next:
    if(pg_tok != PG_IDENT){
      addError(NULL,"ACC: empty name list in ACC directive clause");
	return NULL;
    }
    args = exprListAdd(args, pg_tok_val);
    pg_get_token();
    if(pg_tok == ','){
	pg_get_token();
	goto next;
    } else if(pg_tok == ')'){
	pg_get_token();
	return args;
    } 

    addError(NULL,"ACC: syntax error in ACC pragma clause");
    return NULL;
}

static CExpr* parse_ACC_reduction_namelist(int *r)
{
  CExpr* args;

  args = EMPTY_LIST;
    if(pg_tok != '('){
      addError(NULL,"ACC reduction clause requires name list");
	return NULL;
    }
    pg_get_token();
    switch(pg_tok){
    case '+': *r = ACC_REDUCTION_PLUS; break;
    case '-': *r = ACC_REDUCTION_MINUS; break;
    case '*': *r = ACC_REDUCTION_MUL; break;
    case '&': *r = ACC_REDUCTION_BITAND; break;
    case '|': *r = ACC_REDUCTION_BITOR; break;
    case '^': *r = ACC_REDUCTION_BITXOR; break;
    case PG_ANDAND: *r = ACC_REDUCTION_LOGAND; break;
    case PG_OROR: *r = ACC_REDUCTION_LOGOR; break;
    case PG_IDENT:
	if(PG_IS_IDENT("max")) { *r = ACC_REDUCTION_MAX; break; }
	if(PG_IS_IDENT("min")) { *r = ACC_REDUCTION_MIN; break; }
    default:
	return NULL;	/* syntax error */
    }
    pg_get_token();
    if(pg_tok != ':') return NULL;
    pg_get_token();

 next:
    if(pg_tok != PG_IDENT){
      addError(NULL,"empty name list in ACC reduction clause");
	return NULL;
    }
    args = exprListAdd(args,pg_tok_val);
    pg_get_token();
    if(pg_tok == ','){
	pg_get_token();
	goto next;
    } else if(pg_tok == ')'){
	pg_get_token();
	return args;
    } 

    addError(NULL,"syntax error in ACC directive clause");
    return NULL;
}

static CExpr* parse_ACC_clause_arg()
{
    CExpr *v;

    if(pg_tok != '('){
      addError(NULL,"ACC reduction clause requires argument");
      return NULL;
    }
    pg_get_token();
    if((v = pg_parse_expr()) == NULL) goto syntax_err;
    if(pg_tok != ')') goto syntax_err;
    pg_get_token();
    return v;

  syntax_err:
    addError(NULL,"ACC: syntax error in ACC pragma clause");
    return NULL;
}

